/****************************************************************************
 *  SPI flash driver for M25P64
 ****************************************************************************/
#include <common.h>
#include <linux/ctype.h>

#if defined(CONFIG_SPI)

//Application definitions

#define	NUM_SECTORS 	128	/* number of sectors */
#define SECTOR_SIZE	0x10000
#define FLASH_PAGE_SIZE	0x100
#define NOP_NUM		1000

#define COMMON_SPI_SETTINGS (SPE|MSTR|CPHA|CPOL) //Settings to the SPI_CTL
#define TIMOD01 (0x01)		//stes the SPI to work with core instructions
#define BAUD_RATE_DIVISOR	CONFIG_SPI_BAUD

//Flash commands
#define SPI_WREN            (0x06)  //Set Write Enable Latch
#define SPI_WRDI            (0x04)  //Reset Write Enable Latch
#define SPI_RDSR            (0x05)  //Read Status Register
#define SPI_WRSR            (0x01)  //Write Status Register
#define SPI_READ            (0x03)  //Read data from memory
#define SPI_PP              (0x02)  //Program Data into memory
#define SPI_SE              (0xD8)  //Erase one sector in memory
#define SPI_BE              (0xC7)  //Erase all memory
#define WIP					(0x1)	//Check the write in progress bit of the SPI status register
#define WEL					(0x2)	//Check the write enable bit of the SPI status register

#define TIMEOUT 350000000

typedef enum
{
	NO_ERR,
	POLL_TIMEOUT,
	INVALID_SECTOR,
	INVALID_BLOCK,
}ERROR_CODE;


/* -------------------
 * Variables
 * ------------------- */

static inline void Send_Byte(unsigned char ch)
{
	unsigned short dummyread;

	*pSPI_TDBR = ch;
	 __builtin_bfin_ssync();
	while( (*pSPI_STAT & TXS) != 0);
	while((*pSPI_STAT & SPIF) == 0);
	while((*pSPI_STAT & RXS) == 0);
	dummyread = *pSPI_RDBR;			// Read dummy to empty the receive register

}

static void SetupSPI( const int spi_setting )
{

	if(icache_status()||dcache_status())
                udelay(CONFIG_CCLK_HZ/50000000);
	/*sets up the PF2 to be the slave select of the SPI */
	*pSPI_FLG = 0xFB04;
	*pSPI_BAUD = BAUD_RATE_DIVISOR;
	*pSPI_CTL = spi_setting;
	 __builtin_bfin_ssync();
	udelay(CONFIG_CCLK_HZ/50000000);
}

static void ReleaseSPI(void)
{

//	*pSPI_CTL = 0x0400;	/* disable SPI*/
	*pSPI_CTL = 0xff00;	/* disable SPI*/
	*pSPI_FLG = 0;
	*pSPI_BAUD = 0;
	 __builtin_bfin_ssync();
	udelay(CONFIG_CCLK_HZ/50000000);

}

static void SPI_ON(void)
{
	/*sets up the PF2 to be the slave select of the SPI */
	*pSPI_FLG = 0xFB04;
	 __builtin_bfin_ssync();
	udelay(CONFIG_CCLK_HZ/50000000);
}

static void SPI_OFF(void)
{
	/*sets up the PF2 to be the slave select of the SPI */
	*pSPI_FLG = 0xFF04;
	 __builtin_bfin_ssync();
	udelay(CONFIG_CCLK_HZ/50000000);
}


static void SendSingleCommand( const int iCommand )
{
	/*turns on the SPI in single write mode*/
	SPI_ON();

	/* sends the actual command to the SPI TX register*/
	Send_Byte(iCommand);

	/*The SPI will be turned off*/
	SPI_OFF();

}

static inline void Send_Long_Command(unsigned char ch,
				     unsigned long addr)
{
	/*turns on the SPI in single write mode*/
	SPI_ON();

	Send_Byte(ch);
	// Send Command
	Send_Byte(addr >> 16);
	// Send the highest byte of the 24 bit address first
	Send_Byte((addr >> 8) & 0X0000FF);
	// Send the middle byte of the 24 bit address  second
	Send_Byte(addr & 0X0000FF);
	// Send the lowest byte of the 24 bit address last
}

static char ReadStatusRegister(void)
{
	char status_register;

	/*turns on the SPI in single write mode*/
	SPI_ON();

	Send_Byte(SPI_RDSR);	/* send instruction to read status register */
	*pSPI_TDBR = 0;		/*send dummy to receive the status register*/
	 __builtin_bfin_ssync();
	while((*pSPI_STAT & TXS) != 0);
	while((*pSPI_STAT & SPIF) == 0);
	while((*pSPI_STAT & RXS) == 0);
	status_register = *pSPI_RDBR;	/* read the status register*/

	SPI_OFF();		/* Turn off the SPI */

	return status_register;
}

static ERROR_CODE Wait_For_WEL(void)
{
	int i;
	char status_register;

	for (i = 0; i < TIMEOUT; i++) {
		status_register = ReadStatusRegister();
		if ((status_register & WEL) != 0)
			return NO_ERR;
	};
	printf("Timeout in Wait_For_WEL\n");
	return POLL_TIMEOUT;
}

static ERROR_CODE Wait_For_WIP_Zero(void)
{
	int i;
	char status_register;

	for (i = 0; i < TIMEOUT; i++) {
		status_register = ReadStatusRegister();
		if( (status_register & WIP) == 0)
			return NO_ERR;
	};
	printf("Timeout in Wait_For_WIP_Zero\n");
	return POLL_TIMEOUT;
}


static ERROR_CODE GetSectorNumber(unsigned long ulOffset, int *pnSector )
{
	int nSector = 0;
	ERROR_CODE ErrorCode = NO_ERR;

	if(ulOffset > (NUM_SECTORS * SECTOR_SIZE - 1)){
		ErrorCode = INVALID_SECTOR;
		return ErrorCode;
	}

	nSector = (int)ulOffset / SECTOR_SIZE;
	*pnSector = nSector;

	// ok
	return ErrorCode;
}

static ERROR_CODE EraseBlock( int nBlock )
{
	unsigned long ulSectorOff;
	ERROR_CODE ErrorCode = NO_ERR;

	// if the block is invalid just return
	if ( (nBlock < 0) || (nBlock > NUM_SECTORS) )
	{
		ErrorCode = INVALID_BLOCK;	// tells us if there was an error erasing flash
		return ErrorCode;
	}

	ulSectorOff = (nBlock * SECTOR_SIZE);

	// A write enable instruction must previously have been executed
	SendSingleCommand(SPI_WREN);

	//The status register will be polled to check the write enable latch "WREN"
	ErrorCode = Wait_For_WEL();

	if( POLL_TIMEOUT == ErrorCode )
	{
		printf("SPI Erase block error\n");
		return ErrorCode;
	}

	// Send the erase block command to the flash followed by the 24 address
	// to point to the start of a sector.
	Send_Long_Command(SPI_SE, ulSectorOff);

	//Turns off the SPI
	SPI_OFF();

	// Poll the status register to check the Write in Progress bit
	// Sector erase takes time
	ErrorCode = Wait_For_WIP_Zero();

	// block erase should be complete
	return ErrorCode;
}

static ERROR_CODE WritePage ( unsigned long ulStartAddr, uchar *temp)
{

	int i;
	ERROR_CODE ErrorCode = NO_ERR;	// tells us if there was an error erasing flash

	// First, a Write Enable Command must be sent to the SPI.
	SendSingleCommand(SPI_WREN);

	// Second, the SPI Status Register will be tested whether the
	//         Write Enable Bit has been set.
	ErrorCode = Wait_For_WEL();
	if( POLL_TIMEOUT == ErrorCode ) {
		printf("SPI Write Time Out\n");
		return ErrorCode;
	}
	// Third, the 24 bit address will be shifted out the SPI MOSI bytewise.
	Send_Long_Command(SPI_PP, ulStartAddr);
	// Fourth, maximum number of 256 bytes will be taken from the Buffer
	// and sent to the SPI device.
	for (i = 0; i < FLASH_PAGE_SIZE; i++) {
		Send_Byte(*temp++);
	}

	SPI_OFF(); // Turns the SPI off

	// Sixth, the SPI Write in Progress Bit must be toggled to ensure the
	// programming is done before start of next transfer.
	ErrorCode = Wait_For_WIP_Zero();

	if( POLL_TIMEOUT == ErrorCode )
		printf("SPI Program Time out!\n");

	return ErrorCode;
}

static ERROR_CODE WriteSector( unsigned long ulStart, uchar *pnData )
{

	int lwCount;

	ERROR_CODE ErrorCode = NO_ERR;

	lwCount = SECTOR_SIZE;

	while (lwCount != 0)
	{
		ErrorCode = WritePage(ulStart, pnData);

		// After each function call of WritePage the counter must be adjusted
		lwCount -= FLASH_PAGE_SIZE;

		// Also, both address pointers must be recalculated.
		ulStart += FLASH_PAGE_SIZE;
		pnData  += FLASH_PAGE_SIZE;
	}

	// return the appropriate error code
	return ErrorCode;
}

/*****************************************************************************
* ERROR_CODE ReadData()
*
* Read a value from flash
*
* Inputs:	unsigned long ulStart - holds the SPI start address
*			 long lCount - number of elements to read
*			 char * cnData - pointer to store value read from flash
***************************************************************************** */
static ERROR_CODE ReadData(unsigned long ulStart, long lCount, uchar * cnData)
{
	int i;

	Send_Long_Command(SPI_READ, ulStart);	// Send the read command to SPI device

	// After the SPI device address has been placed on the MOSI pin the data can be
	// received on the MISO pin.
	for (i = 0; i < lCount; i++) {
		*pSPI_TDBR = 0;			//send dummy
		 __builtin_bfin_ssync();
		while((*pSPI_STAT & RXS) == 0);
		*cnData++  = *pSPI_RDBR;	//read
	}
	SPI_OFF();				// Turn off the SPI

	return NO_ERR;
}

/****************************************************************************
 *  Function:    spi_read
 **************************************************************************** */
ssize_t spi_read (uchar *addr, int alen, uchar *buffer, int len)
{
	unsigned long offset;

	SetupSPI( (COMMON_SPI_SETTINGS|TIMOD01) );

	offset = (((addr[0] << 8) | addr[1]) << 8) | addr[2];
	ReadData ( offset, len, buffer );
	ReleaseSPI();
	return len;
}

/****************************************************************************
 *  Function:    spi_write
 **************************************************************************** */
ssize_t spi_write (uchar *addr, int alen, uchar *buffer, int len)
{
	unsigned long offset;
	int	start_block,end_block;
	int start_byte,end_byte;
	ERROR_CODE result = NO_ERR;
	uchar temp [SECTOR_SIZE];
	int i,num;

	SetupSPI((COMMON_SPI_SETTINGS|TIMOD01));

	offset = addr[0]<<16 | addr[1] <<8 | addr[2];
	/* Get the start block number */
	if (GetSectorNumber(offset, &start_block) != NO_ERR) {
		printf ("Invalid start offset 0x%08X\n", addr);
		return 0;
	}

	/* Get the end block number */
	if (GetSectorNumber(offset + len - 1, &end_block) != NO_ERR) {
		printf ("Invalid end offset 0x%08X\n", offset + len - 1);
		return 0;
	}

	for(num = start_block; num <= end_block; num ++){
		start_byte = num * SECTOR_SIZE;
		ReadData (start_byte, SECTOR_SIZE, temp);
		end_byte = (num+1) * SECTOR_SIZE -1;
		if(start_byte < offset)
			start_byte = offset;
		if( end_byte >= (offset+len))
			end_byte = (offset+len-1);
		for (i = start_byte; i <= end_byte; i++)
			temp [i - num * SECTOR_SIZE] = buffer [i - offset];
		EraseBlock (num);
		result = WriteSector (num * SECTOR_SIZE, temp);
		if (result != NO_ERR)
			return 0;
		printf("\r0x%08X", start_byte);
	}
	printf("\r0x%08X", offset + len - 1);
//	printf("\n");
	ReleaseSPI();
	return len;
}

/* **************************************************************************
 *
 *  Function:    spi_init_f
 *
 *  Description: Init SPI-Controller (ROM part)
 *
 *  return:      ---
 *
 * *********************************************************************** */
void spi_init_f (void)
{
}

/* **************************************************************************
 *
 *  Function:    spi_init_r
 *
 *  Description: Init SPI-Controller (RAM part) -
 *		 The malloc engine is ready and we can move our buffers to
 *		 normal RAM
 *
 *  return:      ---
 *
 * *********************************************************************** */
void spi_init_r (void)
{
	return;
}
#endif	/* CONFIG_SPI */
